@@ -1,5 +1,5 @@
 /* Target file hash table management for GNU Make.
-Copyright (C) 1988,89,90,91,92,93,94,95,96 Free Software Foundation, Inc.
+Copyright (C) 1988,89,90,91,92,93,94,95,96,97 Free Software Foundation, Inc.
 This file is part of GNU Make.
 
 GNU Make is free software; you can redistribute it and/or modify
@@ -49,6 +49,9 @@ lookup_file (name)
   register struct file *f;
   register char *n;
   register unsigned int hashval;
+#ifdef VMS
+  register char *lname, *ln;
+#endif
 
   if (*name == '\0')
     abort ();
@@ -57,6 +60,12 @@ lookup_file (name)
      for names read from makefiles.  It is here for names passed
      on the command line.  */
 #ifdef VMS
+  lname = (char *)malloc(strlen(name) + 1);
+  for (n=name, ln=lname; *n != '\0'; ++n, ++ln)
+    *ln = isupper(*n) ? tolower(*n) : *n;
+  *ln = '\0';
+  name = lname;
+
   while (name[0] == '[' && name[1] == ']' && name[2] != '\0')
       name += 2;
 #endif
@@ -87,11 +96,17 @@ lookup_file (name)
 
   for (f = files[hashval]; f != 0; f = f->next)
     {
-      if (strieq (f->name, name))
+      if (strieq (f->hname, name))
 	{
+#ifdef VMS
+	  free (lname);
+#endif
 	  return f;
 	}
     }
+#ifdef VMS
+  free (lname);
+#endif
   return 0;
 }
 
@@ -128,7 +143,7 @@ enter_file (name)
   hashval %= FILE_BUCKETS;
 
   for (f = files[hashval]; f != 0; f = f->next)
-    if (strieq (f->name, name))
+    if (strieq (f->hname, name))
       break;
 
   if (f != 0 && !f->double_colon)
@@ -141,7 +156,7 @@ enter_file (name)
 
   new = (struct file *) xmalloc (sizeof (struct file));
   bzero ((char *) new, sizeof (struct file));
-  new->name = name;
+  new->name = new->hname = name;
   new->update_status = -1;
 
   if (f == 0)
@@ -162,16 +177,16 @@ enter_file (name)
   return new;
 }
 
-/* Rename FILE to NAME.  This is not as simple as resetting
-   the `name' member, since it must be put in a new hash bucket,
+/* Rehash FILE to NAME.  This is not as simple as resetting
+   the `hname' member, since it must be put in a new hash bucket,
    and possibly merged with an existing file called NAME.  */
 
 void
-rename_file (file, name)
+rehash_file (file, name)
      register struct file *file;
      char *name;
 {
-  char *oldname = file->name;
+  char *oldname = file->hname;
   register unsigned int oldhash;
   register char *n;
 
@@ -187,6 +202,23 @@ rename_file (file, name)
   file_hash_enter (file, name, oldhash, file->name);
 }
 
+/* Rename FILE to NAME.  This is not as simple as resetting
+   the `name' member, since it must be put in a new hash bucket,
+   and possibly merged with an existing file called NAME.  */
+
+void
+rename_file (file, name)
+     register struct file *file;
+     char *name;
+{
+  rehash_file(file, name);
+  while (file)
+    {
+      file->name = file->hname;
+      file = file->prev;
+    }
+}
+
 void
 file_hash_enter (file, name, oldhash, oldname)
      register struct file *file;
@@ -208,9 +240,12 @@ file_hash_enter (file, name, oldhash, oldname)
   /* Look for an existing file under the new name.  */
 
   for (oldfile = files[newbucket]; oldfile != 0; oldfile = oldfile->next)
-    if (strieq (oldfile->name, name))
+    if (strieq (oldfile->hname, name))
       break;
 
+  /* If the old file is the same as the new file, something's wrong.  */
+  assert (oldfile != file);
+
   if (oldhash != 0 && (newbucket != oldbucket || oldfile != 0))
     {
       /* Remove FILE from its hash bucket.  */
@@ -228,9 +263,9 @@ file_hash_enter (file, name, oldhash, oldname)
 
   /* Give FILE its new name.  */
 
-  file->name = name;
+  file->hname = name;
   for (f = file->double_colon; f != 0; f = f->prev)
-    f->name = name;
+    f->hname = name;
 
   if (oldfile == 0)
     {
@@ -319,6 +354,7 @@ in favor of those for `%s'.",
       MERGE (is_target);
       MERGE (cmd_target);
       MERGE (phony);
+      MERGE (ignore_vpath);
 #undef MERGE
 
       file->renamed = oldfile;
@@ -353,9 +389,9 @@ remove_intermediates (sig)
 	  if (f->update_status == -1)
 	    /* If nothing would have created this file yet,
 	       don't print an "rm" command for it.  */
-	    continue;
-	  else if (just_print_flag)
-	    status = 0;
+            continue;
+ 	  else if (just_print_flag)
+  	    status = 0;
 	  else
 	    {
 	      status = unlink (f->name);
@@ -520,9 +556,7 @@ print_file (f)
      struct file *f;
 {
   register struct dep *d;
-#ifdef VMS
-  extern char *cvt_time PARAMS ((unsigned long));
-#endif
+
   putchar ('\n');
   if (!f->is_target)
     puts ("# Not a target:");
@@ -558,13 +592,8 @@ print_file (f)
   else if (f->last_mtime == (time_t) -1)
     puts ("#  File does not exist.");
   else
-#ifdef VMS
-    printf ("#  Last modified %.24s (%0lx)\n",
-	    cvt_time(f->last_mtime), (unsigned long) f->last_mtime);
-#else
     printf ("#  Last modified %.24s (%ld)\n",
 	    ctime (&f->last_mtime), (long int) f->last_mtime);
-#endif
   printf ("#  File has%s been updated.\n",
 	  f->updated ? "" : " not");
   switch (f->command_state)
@@ -646,8 +675,8 @@ print_file_data_base ()
     {
       printf ("\n# %u files in %u hash buckets.\n", nfiles, FILE_BUCKETS);
 #ifndef	NO_FLOAT
-      printf ("# average %.1f files per bucket, max %u files in one bucket.\n",
-	      ((double) nfiles) / ((double) FILE_BUCKETS) * 100.0, per_bucket);
+      printf ("# average %.3f files per bucket, max %u files in one bucket.\n",
+	      ((double) nfiles) / ((double) FILE_BUCKETS), per_bucket);
 #endif
     }
 }
