@@ -30,6 +30,15 @@ struct rule
     char in_use;		/* If in use by a parent pattern_search.  */
   };
 
+struct pattern_var
+  {
+    struct pattern_var *next;
+    char *target;
+    unsigned int len;
+    char *suffix;
+    struct variable_set_list *vars;
+  };
+
 /* For calling install_pattern_rule.  */
 struct pspec
   {
@@ -51,3 +60,5 @@ extern unsigned int maxsuffix;
 
 extern void install_pattern_rule PARAMS ((struct pspec *p, int terminal));
 extern int new_pattern_rule PARAMS ((struct rule *rule, int override));
+extern struct pattern_var *create_pattern_var PARAMS ((char *target, char *suffix));
+extern struct pattern_var *lookup_pattern_var PARAMS ((char *target));
