@@ -6,6 +6,7 @@
 #include "sub_proc.h"
 #include "proc.h"
 #include "w32err.h"
+#include "config.h"
 
 static char *make_command_line( char *shell_name, char *exec_path, char **argv);
 
@@ -319,14 +320,10 @@ find_file(char *exec_path, LPOFSTRUCT file_info)
 	char *fname;
 	char *ext;
 
-	if ((exec_handle = (HANDLE)OpenFile(exec_path, file_info,
-			OF_READ | OF_SHARE_COMPAT)) != (HANDLE)HFILE_ERROR) {
-		return(exec_handle);
-	}
-
 	fname = malloc(strlen(exec_path) + 5);
 	strcpy(fname, exec_path);
 	ext = fname + strlen(fname);
+
 	strcpy(ext, ".exe");
 	if ((exec_handle = (HANDLE)OpenFile(fname, file_info,
 			OF_READ | OF_SHARE_COMPAT)) != (HANDLE)HFILE_ERROR) {
@@ -334,6 +331,13 @@ find_file(char *exec_path, LPOFSTRUCT file_info)
 		return(exec_handle);
 	}
 
+	strcpy(ext, ".cmd");
+	if ((exec_handle = (HANDLE)OpenFile(fname, file_info,
+			OF_READ | OF_SHARE_COMPAT)) != (HANDLE)HFILE_ERROR) {
+		free(fname);
+		return(exec_handle);
+	}
+
 	strcpy(ext, ".bat");
 	if ((exec_handle = (HANDLE)OpenFile(fname, file_info,
 			OF_READ | OF_SHARE_COMPAT)) != (HANDLE)HFILE_ERROR) {
@@ -341,6 +345,13 @@ find_file(char *exec_path, LPOFSTRUCT file_info)
 		return(exec_handle);
 	}
 
+	/* should .com come before this case? */
+	if ((exec_handle = (HANDLE)OpenFile(exec_path, file_info,
+			OF_READ | OF_SHARE_COMPAT)) != (HANDLE)HFILE_ERROR) {
+		free(fname);
+		return(exec_handle);
+	}
+
 	strcpy(ext, ".com");
 	if ((exec_handle = (HANDLE)OpenFile(fname, file_info,
 			OF_READ | OF_SHARE_COMPAT)) != (HANDLE)HFILE_ERROR) {
@@ -395,7 +406,7 @@ process_begin(
 	exec_handle = find_file(exec_path, &file_info);
 
 	/*
-	 * If we couldn't open the file, just assume that Win32 will be able
+	 * If we couldn't open the file, just assume that Windows32 will be able
 	 * to find and execute it.
 	 */
 	if (exec_handle == (HANDLE)HFILE_ERROR) {
@@ -489,14 +500,8 @@ process_begin(
 	startInfo.hStdOutput = (HANDLE)pproc->sv_stdout[1];
 	startInfo.hStdError = (HANDLE)pproc->sv_stderr[1];
 
-	/*
-	 * See if we need to setuid to a different user.
-	 */
-	if (as_user) {
-		return -1;
-	}
-
 	if (as_user) {
+		if (envblk) free(envblk);
 		return -1;
 	} else {
 		if (CreateProcess(
@@ -514,6 +519,7 @@ process_begin(
 			pproc->last_err = GetLastError();
 			pproc->lerrno = E_FORK;
 			fprintf(stderr, "process_begin: CreateProcess(%s, %s, ...) failed.\n", exec_path, command_line);
+			if (envblk) free(envblk);
 			free( command_line );
 			return(-1);
 		}
@@ -538,6 +544,7 @@ process_begin(
 	}
 
 	free( command_line );
+	if (envblk) free(envblk);
 	pproc->lerrno=0;
 	return 0;
 }
@@ -578,7 +585,7 @@ proc_stdout_thread(sub_process *pproc)
 	for (;;) {
 		if (ReadFile( (HANDLE)pproc->sv_stdout[0], &c, 1, &nread, NULL) 
 					== FALSE) {
-/*			map_win32_error_to_string(GetLastError());*/
+/*			map_windows32_error_to_string(GetLastError());*/
 			_endthreadex(0);
 		}
 		if (nread == 0)
@@ -609,7 +616,7 @@ proc_stderr_thread(sub_process *pproc)
 
 	for (;;) {
 		if (ReadFile( (HANDLE)pproc->sv_stderr[0], &c, 1, &nread, NULL) == FALSE) {
-			map_win32_error_to_string(GetLastError());
+			map_windows32_error_to_string(GetLastError());
 			_endthreadex(0);
 		}
 		if (nread == 0)
@@ -718,7 +725,7 @@ process_pipe_io(
 			 	one second to collect all remaining output */
 		
 		if (wait_return == WAIT_FAILED) {
-/*			map_win32_error_to_string(GetLastError());*/
+/*			map_windows32_error_to_string(GetLastError());*/
 			pproc->last_err = GetLastError();
 			pproc->lerrno = E_SCALL;
 			goto done;
@@ -835,7 +842,7 @@ process_file_io(
 	wait_return = WaitForSingleObject(childhand, INFINITE);
 		
 	if (wait_return != WAIT_OBJECT_0) {
-/*		map_win32_error_to_string(GetLastError());*/
+/*		map_windows32_error_to_string(GetLastError());*/
 		pproc->last_err = GetLastError();
 		pproc->lerrno = E_SCALL;
 		goto done2;
@@ -882,108 +889,6 @@ process_cleanup(
 }
 
 
-/*
- * Try to protect against WIN32 argument munging. This function takes
- * an argv vector and outputs a 'protected' string as a return
- * value. The return code can be safely passed to CreateProcess().
- *
- * The caller should free the return value.
- */
-
-#define TRACE(x)
-static char *fix_command_line(char *args[])
-{
-	int i;
-	char *narg;
-	char *nargp;
-	char *p;
-	char *q;
-	int alloc_len = 0;
-
-	for (i = 0; args[i]; i++)
-		alloc_len += ((strlen(args[i]) * 2) + 1);
-	/* account for possible enclosing quotes and null termination */
-	alloc_len += 3;
-
-	nargp = narg = malloc(alloc_len);
-
-	for (i = 0; args[i]; i++) {
-		p = args[i];
-		TRACE(("original arg: %s\n", p));
-
-		if (*p == '\0') {
-			*nargp++ = '"';
-			*nargp++ = '"';
-			*nargp = '\0';
-			TRACE(("empty string arg: %s\n", nargp-2));
-		} else if (strpbrk(p, "\" \t")) {
-			/* point to end of copy buffer */
-			q = narg;
-			q += (alloc_len-1);
-			*q-- = '\0'; /* ensure null terminated string */
-			*q-- = '"';  /* terminating quote of argument */
-
-			/* point to end of the input string */
-			p = args[i];
-			p += strlen(args[i]);
-			p--;
-
-			/* 
-			 * Because arg is quoted, escape any backslashes 
-			 * that might occur at the end of the string which
-			 * proceed the closing quote.
-			 * Example:
-			 * 	foo c:\
-			 * Becomes:
-			 *	"foo c:\\"
-			 */
-			while (*p == '\\')
-				*q-- = *p--, *q-- = '\\';
-
-			/* copy the string in reverse */
-			while (p >= args[i]) {
-				/* copy the character */
-				*q-- = *p--;
-
-				/* 
-				 * Escape any double quote found. Also escape
-				 * each backslash preceding the double quote.
-				 */
-				if (*(p+1) == '"') {
-					*q-- = '\\';
-					if (p >= args[i] && *p == '\\')
-						while (p >= args[i] && *p == '\\')
-							*q-- = *p--, *q-- = '\\';
-				}
-			}
-
-			/* finish quoting arg, q now points to complete arg */
-			*q = '"';
-
-			/* rejustify */
-			memmove(nargp, q, strlen(q) + 1);
-			TRACE(("arg with white space or doublequotes: %s\n", nargp));
-			nargp += strlen(nargp);
-		} else {
-			/* just copy the argument, no protection needed */
-			strcpy(nargp, args[i]);
-			TRACE(("plain arg: %s\n", nargp));
-			nargp += strlen(nargp);
-		}
-
-		/* separate arguments with spaces (if more args to gather) */
-		if (args[i+1])
-			*nargp++ = ' ';
-		*nargp   = '\0';
-	} /* end for */
-
-	/* NULL terminate the arg list */
-	*nargp = '\0';
-
-	return (narg);
-}
-#undef TRACE
-
 /*
  * Description: 
  *	 Create a command line buffer to pass to CreateProcess
@@ -998,30 +903,211 @@ static char *fix_command_line(char *args[])
  */
 
 static char *
-make_command_line( char *shell_name, char *exec_path, char **argv)
+make_command_line( char *shell_name, char *full_exec_path, char **argv)
 {
-	char** nargv;
-	char*  buf;
-	int    i;
- 
-	if (shell_name) {
-		for (i = 0; argv[i]; i++);
-		i += 2;
-		nargv = (char **) malloc(i * sizeof (char *));
-		nargv[0] = shell_name;
-		for (i = 1; argv[i-1]; i++)
-			nargv[i] = argv[i-1];
-		nargv[i] = NULL;
-	} else
-		nargv = argv;
+	int		argc = 0;
+	char**		argvi;
+	int*		enclose_in_quotes = NULL;
+	int*		enclose_in_quotes_i;
+	unsigned int	bytes_required = 0;
+	char*		command_line;
+	char*		command_line_i;
 
-	/* create string suitable for CreateProcess() */
-	buf = fix_command_line(nargv);
+	if (shell_name && full_exec_path) {
+		bytes_required
+		  = strlen(shell_name) + 1 + strlen(full_exec_path);
+		/*
+		 * Skip argv[0] if any, when shell_name is given.
+		 */
+		if (*argv) argv++;
+		/*
+		 * Add one for the intervening space.
+		 */
+		if (*argv) bytes_required++;
+	}
 
-	if (shell_name)
-		free(nargv);
-	
-	return buf;
+	argvi = argv;
+	while (*(argvi++)) argc++;
+
+	if (argc) {
+		enclose_in_quotes = (int*) calloc(1, argc * sizeof(int));
+
+		if (!enclose_in_quotes) {
+			return NULL;
+		}
+	}
+
+	/* We have to make one pass through each argv[i] to see if we need
+	 * to enclose it in ", so we might as well figure out how much
+	 * memory we'll need on the same pass.
+	 */
+
+	argvi = argv;
+	enclose_in_quotes_i = enclose_in_quotes;
+	while(*argvi) {
+		char* p = *argvi;
+		unsigned int backslash_count = 0;
+
+		/*
+		 * We have to enclose empty arguments in ".
+		 */
+		if (!(*p)) *enclose_in_quotes_i = 1;
+
+		while(*p) {
+			switch (*p) {
+			case '\"':
+				/*
+				 * We have to insert a backslash for each "
+				 * and each \ that precedes the ".
+				 */
+				bytes_required += (backslash_count + 1);
+				backslash_count = 0;
+				break;
+
+			case '\\':
+				backslash_count++;
+				break;
+	/*
+	 * At one time we set *enclose_in_quotes_i for '*' or '?' to suppress
+	 * wildcard expansion in programs linked with MSVC's SETARGV.OBJ so
+	 * that argv in always equals argv out. This was removed.  Say you have
+	 * such a program named glob.exe.  You enter
+	 * glob '*'
+	 * at the sh command prompt.  Obviously the intent is to make glob do the
+	 * wildcarding instead of sh.  If we set *enclose_in_quotes_i for '*' or '?',
+	 * then the command line that glob would see would be
+	 * glob "*"
+	 * and the _setargv in SETARGV.OBJ would _not_ expand the *.
+	 */
+			case ' ':
+			case '\t':
+				*enclose_in_quotes_i = 1;
+				/* fall through */
+
+			default:
+				backslash_count = 0;
+				break;
+			}
+			
+			/*
+			 * Add one for each character in argv[i].
+			 */
+			bytes_required++;
+
+			p++;
+		}
+
+		if (*enclose_in_quotes_i) {
+			/*
+			 * Add one for each enclosing ",
+			 * and one for each \ that precedes the
+			 * closing ".
+			 */
+			bytes_required += (backslash_count + 2);
+		}
+		
+		/*
+		 * Add one for the intervening space.
+		 */
+		if (*(++argvi)) bytes_required++;
+		enclose_in_quotes_i++;
+	}
+
+	/*
+	 * Add one for the terminating NULL.
+	 */
+	bytes_required++;
+
+	command_line = (char*) malloc(bytes_required);
+
+	if (!command_line) {
+		if (enclose_in_quotes) free(enclose_in_quotes);
+		return NULL;
+	}
+
+	command_line_i = command_line;
+
+	if (shell_name && full_exec_path) {
+		while(*shell_name) {
+			*(command_line_i++) = *(shell_name++);
+		}
+
+		*(command_line_i++) = ' ';
+
+		while(*full_exec_path) {
+			*(command_line_i++) = *(full_exec_path++);
+		}
+
+		if (*argv) {
+			*(command_line_i++) = ' ';
+		}
+	}
+
+	argvi = argv;
+	enclose_in_quotes_i = enclose_in_quotes;
+
+	while(*argvi) {
+		char* p = *argvi;
+		unsigned int backslash_count = 0;
+
+		if (*enclose_in_quotes_i) {
+			*(command_line_i++) = '\"';
+		}
+
+		while(*p) {
+			if (*p == '\"') {
+				/*
+				 * We have to insert a backslash for the "
+				 * and each \ that precedes the ".
+				 */
+				backslash_count++;
+
+				while(backslash_count) {
+					*(command_line_i++) = '\\';
+					backslash_count--;
+				};
+
+			} else if (*p == '\\') {
+				backslash_count++;
+			} else {
+				backslash_count = 0;
+			}
+
+			/*
+			 * Copy the character.
+			 */
+			*(command_line_i++) = *(p++);
+		}
+
+		if (*enclose_in_quotes_i) {
+			/*
+			 * Add one \ for each \ that precedes the
+			 * closing ".
+			 */
+			while(backslash_count--) {
+				*(command_line_i++) = '\\';
+			};
+				
+			*(command_line_i++) = '\"';
+		}
+		
+		/*
+		 * Append an intervening space.
+		 */
+		if (*(++argvi)) {
+			*(command_line_i++) = ' ';
+		}
+		
+		enclose_in_quotes_i++;
+	}
+
+	/*
+	 * Append the terminating NULL.
+	 */
+	*command_line_i = '\0';
+
+	if (enclose_in_quotes) free(enclose_in_quotes);
+	return command_line;
 }
 
 /*
